class Position:
    """Holds information about a position in the portfolio."""

    def __init__(self, ticker, qty, price, target_allocation=None):
        self._ticker = ticker
        self._qty = qty
        self._price = price
        self._target_allocation = target_allocation

    @property
    def ticker(self):
        return self._ticker

    @property
    def qty(self):
        return self._qty

    @property
    def price(self):
        return self._price

    @property
    def target_allocation(self):
        return self._target_allocation

    @qty.setter
    def qty(self, qty):
        self._qty = qty

    @target_allocation.setter
    def target_allocation(self, target_allocation):
        self._target_allocation = target_allocation



import click
import copy


class Rebalancer:
    """Prints a list of buys that will help the current portfolio move closest
    to the target allocations.
    """

    @staticmethod
    def _position_with_highest_drift(portfolio):
        """Finds the position within the portfolio with the highest drift
        percentage. Positions that can not be bought because the price of
        acquiring a single share is more than the remaining cash amount are not
        considered.
        """
        def valid_drift_pct(drift_pcts):
            ticker = drift_pcts[0]
            position_pct = drift_pcts[1]
            price = portfolio[ticker].price
            return price <= portfolio.buying_power and position_pct <= 0.0
        position_with_highest_drift = None
        highest_drift_pct = None
        drift_percentages = portfolio.drift_percentages()
        # filter out positions where the current allocation is bigger than the
        # target allocation since a buy will only make this drift bigger, and
        # where the price to acquire a single share of the asset is bigger than
        # the remaining cash available
        drift_percentages = dict(filter(valid_drift_pct,
                                        drift_percentages.items()))
        for ticker, drift_pct in drift_percentages.items():
            position = portfolio[ticker]
            if not position_with_highest_drift:
                position_with_highest_drift = position
                highest_drift_pct = drift_pct
            elif drift_pct < highest_drift_pct:
                position_with_highest_drift = position
                highest_drift_pct = drift_pct
            elif (drift_pct == highest_drift_pct and
                  position.price < position_with_highest_drift.price):
                position_with_highest_drift = position
                highest_drift_pct = drift_pct
        return position_with_highest_drift

    @staticmethod
    def print_buys_for_rebalancing(old_portfolio):
        """Prints a list of buys needed to bring the portfolio as close as
        possible to the target allocations.
        """
        new_portfolio = copy.deepcopy(old_portfolio)
        buys = {}
        while new_portfolio.buying_power > 0.0:
            position = Rebalancer._position_with_highest_drift(new_portfolio)
            if not position:
                # No positions remain that are below the target allocation and
                # have a unit price less than the remaning cash amount
                break
            new_portfolio[position.ticker].qty += 1
            new_portfolio.buying_power -= position.price
            if position.ticker not in buys:
                buys[position.ticker] = 0
            buys[position.ticker] += 1
        for ticker, buy_amount in buys.items():
            position = new_portfolio[ticker]
            allocation = new_portfolio.get_current_allocation(ticker)
            click.echo("Buy {}X {} @ {:.2f} - New allocation {:.2f}%".format(
                buy_amount, ticker, position.price, round(allocation, 2)))
        click.echo("Remaining cash ${:.2f}".format(new_portfolio.buying_power))



import click

from ws_rebalancer.target_allocations_csv_reader import (
    TargetAllocationsCsvReader
)
from ws_rebalancer.rebalancer import Rebalancer
from ws_rebalancer.wealthsimple_login import WealthSimpleLogin
from ws_rebalancer.wealthsimple_portfolio_reader import (
    WealthSimplePortfolioReader
)


@click.group()
def ws_rebalancer():
    pass


@ws_rebalancer.command(help=(
    "Generate the recommended buys to get your portfolio as close as possible "
    "to your target allocation."))
@click.option('-t', '--target-allocations-csv', 'target_allocations_csv',
              required=True,
              help="CSV-file containing the target allocations for each "
                   "ticker")
@click.option('--email', required=True, help="Email for WealthSimple login")
@click.password_option(help="Password for WealthSimple login", required=True)
@click.option('--2fa', 'two_factor_auth', is_flag=True,
              help="Enable this flag if your WealthSimple login requires 2FA")
def rebalance(target_allocations_csv, email, password, two_factor_auth):
    try:
        ws = WealthSimpleLogin(email, password,
                               two_factor_auth=two_factor_auth)
    except Exception as e:
        raise click.ClickException("{}".format(str(e)))
    portfolio = WealthSimplePortfolioReader(ws).get_portfolio()
    TargetAllocationsCsvReader(target_allocations_csv, ws).update_portfolio(
        portfolio)
    Rebalancer.print_buys_for_rebalancing(portfolio)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ws_rebalancer']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0', 'wealthsimple-trade-python>=1.1.0,<2.0.0']

entry_points = \
{'console_scripts': ['ws-rebalancer = ws_rebalancer.cli:ws_rebalancer']}

setup_kwargs = {
    'name': 'ws-rebalancer',
    'version': '1.0.1',
    'description': 'A CLI tool that helps you rebalance your WealthSimple portfolios.',
    'long_description': "WealthSimple Portfolio Rebalancer\n=\n[![Build Status](https://travis-ci.com/EmilMaric/ws-rebalancer.svg?branch=main)](https://travis-ci.com/EmilMaric/ws-rebalancer)\n[![codecov](https://codecov.io/gh/EmilMaric/ws-rebalancer/branch/main/graph/badge.svg?token=XJ371LIRJB)](https://codecov.io/gh/EmilMaric/ws-rebalancer)\n\nA CLI tool that helps you rebalance your WealthSimple portfolio. The tool takes in a CSV-file that contains the target allocations for your\nportfolio, and prints out a list of buys you should make to bring your portfolio closer to the target allocations you specified.\n\nThe tool will log you in to your WealthSimple account, where it will fetch your current positions & buying power, and will parse the CSV-file you defined to determine how far away each position is from the target allocation. It will then select the positions that are the farthest away from their target positions iteratively, until your buying power is used up.\n\n## Limitations\n- Currently the tool only supports doing buy-only rebalancing, meaning __it won't__ try to simulate any sells in order to rebalance.\n- The tool only works with CAD denominated stocks and assumes your buying power is in CAD. If there is appetite for supporting other currencies then I can try to support that as well.\n- Prices for the stocks are fetched off the WealthSimple Trade API as well, so the quotes may be delayed by 15 minutes.\n\n# Installation\n```\n# To get the latest release\npip install ws-rebalancer\n```\n\n# CSV-file requirements\nTo start, create a CSV file that will contain tickers from assets in your current portfolio, as well as tickers for any new assets you would like to add.\nIn the CSV file, each line will represent a unique asset you own or want to own. On each line, you need to include the correct ticker for the asset and the target allocation for that asset. The format for each line is as follows:\n```\n<stock ticker>, <target allocation>\n```\n\nHere is a sample CSV file:\n```\n# sample-target-allocations.csv\nMSFT, 50%\nAPPL, 30%\nGOOG, 20%\n```\n\nIn other words, I want my sample portfolio to be 50% Microsoft, 30% Apple, and 20% Google.\n\nMake sure that the target allocations add up to 100%. The tool will raise an error and quit if that is not the case. You can also use decimals to represent fractional target allocations. Also don't repeat the same ticker twice. I've tried my best to sanitize the input and raise any errors that I can forsee with the CSV-file, but there may be some things that I didn't catch so please be careful.\n\n# Usage\nTo generate the buys that will rebalance your portfolio as close as possible to the target allocation, run the tool as follows:\n```\n$ ws-rebalancer rebalance -t <target-allocations-CSV-file> --email <WealthSimple-email-login> [--2fa]\n```\n\nUsing our `sample-target-allocations.csv` as above, a sample run could look as follows:\n```\n$ ws-rebalancer rebalance -t sample-target-allocations.csv --email test@gmail.com --2fa\nPassword:\nRepeat for confirmation:\nEnter 2FA code: 12345\n0. non-registered\nPlease input the account you want: 0\nBuy 5X MSFT @ 10.00 - New allocation 40.00%\nBuy 1X APPL @ 20.00 - New allocation 30.00%\nBuy 1X GOOG @ 30.00 - New allocation 30.00%\nRemaining cash $0.00\n```\nNote that in this example, `MSFT` price was $10.00, `APPL` price was $20.00, and `GOOG` price was $30.00, and we had a buying power of $60.00 in this\naccount (although it is not shown). The tool also fetched our current positions and buying power for the specified account and then provided a list of\nbuys we should make in order to try to meet the specified target allocations.\n\n# Finding Issues\nIf you find issues using this tool, please create an Issue using the [Github issue tracker](https://github.com/EmilMaric/ws-rebalancer/issues)\nand I will try to address it as soon as I can.\n\n# Contributing\nIf you would like to contribute, please read the [CONTRIBUTING.md](https://github.com/EmilMaric/ws-rebalancer/blob/main/CONTRIBUTING.md) page\n",
    'author': 'Emil Maric',
    'author_email': 'emil.maric@hotmail.com',
    'maintainer': 'Emil Maric',
    'maintainer_email': 'emil.maric@hotmail.com',
    'url': 'https://github.com/EmilMaric/ws-rebalancer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



class Security:
    """Contains information about a security on WealthSimple."""

    def __init__(self, security_id, ticker, price):
        self._security_id = security_id
        self._ticker = ticker
        self._price = price

    @property
    def ticker(self):
        return self._ticker

    @property
    def price(self):
        return self._price



import click
import wealthsimple


class WealthSimpleLogin(wealthsimple.WSTrade):
    """Wraps the WealthSimple Trade API with a login interface. This class can
    is used to login to the client's WealthSimple account and interact with the
    WealthSimple Trade API.
    """

    def __init__(self, email, password, two_factor_auth=False):
        two_factor_callback = (
            self.two_factor_function if two_factor_auth else None)
        wealthsimple.WSTrade.__init__(self,
                                      email,
                                      password,
                                      two_factor_callback=two_factor_callback)

    def two_factor_function(self):
        """Retrieves the two-factor auth code from the client."""
        MFACode = ""
        while not MFACode:
            # Obtain user input and ensure it is not empty
            MFACode = click.prompt("Enter 2FA code")
        return MFACode



import click

from ws_rebalancer.portfolio import Portfolio
from ws_rebalancer.position import Position


class WealthSimplePortfolioReader:
    """Reads in the selected WealthSimple account along with the buying power
    and positions in that account and generates a Portfolio object with this
    information to be used for rebalancing.
    """

    def __init__(self, ws):
        self._ws = ws

    def _generate_portfolio(self, buying_power, positions):
        portfolio = Portfolio(buying_power)
        for position in positions:
            ticker = position['stock']['symbol']
            qty = position['quantity']
            price = float(position['quote']['amount'])
            position = Position(ticker, qty, price)
            portfolio.add_position(position)
        return portfolio

    def get_portfolio(self):
        account_ids = self._ws.get_account_ids()
        for account_num, account_id in enumerate(account_ids):
            click.echo("{}. {}".format(account_num, account_id))
        account_idx = click.prompt("Please input the account you want",
                                   type=int)
        account_id = account_ids[account_idx]
        account = self._ws.get_account(account_id)
        buying_power = float(account['buying_power']['amount'])
        positions = self._ws.get_positions(account_id)
        return self._generate_portfolio(buying_power, positions)



import click
import csv

from ws_rebalancer.position import Position
from ws_rebalancer.security import Security


class TargetAllocationsCsvReader:
    '''Reads and verifies the CSV-file containing the following info on each
    row in the following order:
    1. Asset name/ticker
    2. Target allocation for the assert (expressed as a percentage value)

    Will compare the ticker to see if it can be found in your current
    portfolio. If it can be found, it will extract the number of shares owned.
    Otherwise, it will assume that you own 0 shares of the asset.
    '''

    def __init__(self, target_allocations_csv, ws):
        self._target_allocations_csv = target_allocations_csv
        self._ws = ws

    def _verify_columns_in_row(self, row, row_num):
        """Ensure that the number of the columns in the row is expected."""
        if len(row) != 2:
            raise click.ClickException(
                "Row {} malformed - expecting row in this format: "
                "TICKER TARET_ALLOCATION".format(row_num))

    def _verify_all_positions_have_target_allocation(self, portfolio):
        """Ensure that every position in the portfolio has a target
        allocation.
        """
        for position in portfolio.positions.values():
            if not position.target_allocation:
                raise click.ClickException(
                    "Ticker {} does not have a target allocation".format(
                        position.ticker))

    def _verify_total_target_allocation(self, total_target_allocation):
        """Ensure that the total target allocation of all positions in the
        portfolio is exactly 100.
        """
        if total_target_allocation != 100:
            raise click.ClickException(
                "Total combined allocation percentage ({}) of all rows is not "
                "100%".format(total_target_allocation))

    def _sanitize_ticker(self, row):
        """Ticker should be in the first column of the CSV."""
        return row[0]

    def _sanitize_target_allocation(self, row, row_num):
        """Target allocation should be in the second column of the CSV."""
        try:
            target_allocation = row[1].strip('%')
            return float(target_allocation)
        except ValueError:
            raise click.ClickException(
                "Row {} - Target allocation is not a float/integer".format(
                    row_num))

    def _get_security(self, ticker_query, row_num):
        """Search for the security on WealthSimple given a ticker query."""
        securities = self._ws.get_securities_from_ticker(ticker_query)
        if len(securities) == 0:
            raise click.ClickException(
                "Row {} - ticker '{}' cannot be found".format(row_num,
                                                              ticker_query))
        for security_num, security in enumerate(securities):
            ticker = security['stock']['symbol']
            name = security['stock']['name']
            exchange = security['stock']['primary_exchange']
            click.echo("{}. Ticker: {}, Name: {}, Exchange: {}".format(
                security_num, ticker, name, exchange))
        security_idx = click.prompt("Please input the security you want",
                                    type=int)
        security_id = securities[security_idx]['id']
        security = self._ws.get_security(security_id)
        ticker = security['stock']['symbol']
        price = float(security['quote']['amount'])
        return Security(security_id, ticker, price)

    def update_portfolio(self, portfolio):
        """Assign target allocations to the positions in the portfolio and add
        any new positions that are present in the target-allocations CSV but
        not in the current portfolio. These are new positions that we want to
        open in our portfolio.
        """
        seen_tickers = set()
        total_allocation_pct = 0.0
        with open(self._target_allocations_csv, newline='') as f:
            reader = csv.reader(f)
            for row_num, row in enumerate(reader):
                self._verify_columns_in_row(row, row_num)
                ticker = self._sanitize_ticker(row)
                target_allocation = self._sanitize_target_allocation(row,
                                                                     row_num)
                if ticker not in portfolio.positions:
                    # Ticker cannot be found in the current portfolio
                    security = self._get_security(ticker, row_num)
                    if ticker != security.ticker:
                        # Ticker is not spelled correctly in the target
                        # allocations CSV file - needs to be renamed
                        raise click.ClickException(
                            "Row {} - rename ticker '{}' to '{}'".format(
                                row_num, ticker, security.ticker))
                    click.secho(
                        "Warning: '{}' is not in your portfolio".format(
                            ticker),
                        fg='red')
                    position = Position(
                        security.ticker,
                        0,
                        security.price,
                        target_allocation=target_allocation)
                    portfolio.add_position(position)
                else:
                    # Ticker exists in the portfolio
                    position = portfolio[ticker]
                    position.target_allocation = target_allocation
                if ticker in seen_tickers:
                    raise click.ClickException(
                        "Duplicate entry of ticker '{}' on row {}".format(
                            ticker, row_num))
                seen_tickers.add(ticker)
                total_allocation_pct += target_allocation
        self._verify_all_positions_have_target_allocation(portfolio)
        self._verify_total_target_allocation(total_allocation_pct)
        return portfolio



class Portfolio:
    """Represents a collection of positions in a portfolio."""

    def __init__(self, buying_power):
        self._positions = {}
        self._buying_power = buying_power

    def __getitem__(self, ticker):
        return self._positions[ticker]

    def _total(self):
        """The total value of the portfolio."""
        total = 0.0
        for position in self._positions.values():
            total += position.price * position.qty
        return total

    @property
    def positions(self):
        """Get the positions that are a part of this portfolio."""
        return self._positions

    @property
    def buying_power(self):
        """Get the amount of cash that is available to be used for trading."""
        return self._buying_power

    @buying_power.setter
    def buying_power(self, buying_power):
        """Set the new amount of cash that is available to be used for
        trading.
        """
        self._buying_power = buying_power

    def add_position(self, position):
        """Add a position to the portfolio."""
        self._positions[position.ticker] = position

    def drift_percentages(self):
        """Get the drift percentages for all the positions in this portfolio.
        Drift percentage is defined by how far away a position is from its
        target allocation.
        A negative drift percentage means that the current position allocation
        is under the target allocation.
        A positive drift percentage means that the current position allocation
        is over the target allocation.
        """
        drift_pcts = {}
        for position in self._positions.values():
            current_allocation_pct = 0.0
            if self._total() > 0.0:
                position_value = position.price * position.qty
                current_allocation_pct = position_value / self._total()
                current_allocation_pct *= 100
            drift = current_allocation_pct - position.target_allocation
            drift_pct = (drift * 100) / position.target_allocation
            drift_pcts[position.ticker] = drift_pct
        return drift_pcts

    def get_current_allocation(self, ticker):
        """Returns the portion of the current portfolio that is comprised of
        the queried ticker.
        """
        position = self._positions[ticker]
        return (position.price * position.qty * 100) / self._total()
